/*
******************************************************************************

   BLT implementation for X target - stretch functions.

   Copyright (C) 2001	Christoph Egger	[Christoph_Egger@t-online.de]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <math.h>

#include "xblt.h"

#include <ggi/display/xcommon.h>
#include <ggi/display/x.h>

#include <X11/X.h>

#include <ggi/ggi.h>


typedef sint32 fixed;
#define SHIFT	16


int stretch_crossblit(ggiBlt_t src, ggiBlt_t dst,
		    int width, int height,
		    enum ggiGA_rop rop)
{
	struct GA_x_bob *src_bob = NULL;
	struct GA_x_bob *dst_bob = NULL;
	int clipwidth, clipheight;

	ggi_pixel *src_buf;
	ggi_pixel *dst_buf;

	int i;
	int y,ystart;

	fixed src_ystart;
	fixed src_idx, src_y;
	fixed src_xinc, src_yinc;

	ggi_color color;

	if (src == NULL) return GGI_EARGINVAL;
	if (dst == NULL) return GGI_EARGINVAL;

	src_bob = GETBOB(src);
	dst_bob = GETBOB(dst);

	src_buf = src_bob->bobimage;
	dst_buf = dst_bob->bobimage;

#warning Convert width, height and src->w,src->h into the coord base the dst is.


	/* We perform clipping by cutting the outside parts out */
	if (width > dst->w) {
		clipwidth = dst->w;
	} else {
		clipwidth = width;
	}	/* if */

	if (height > dst->h) {
		clipheight = dst->h;
	} else {
		clipheight = height;
	}	/* if */



	ystart = y = 0;

	src_ystart = src_y = 0;
	src_idx = 0;
	src_xinc = ((src->w << SHIFT) / width) + 1;
	src_yinc = ((src->h << SHIFT) / height) + 1;

	for (i = 0; i <= ((width - 1) * (clipheight - 1)); ) {

		ggiUnmapPixel(src->vis, src_buf[(src_idx >> SHIFT)], &color);
		dst_buf[i] = ggiMapColor(dst->vis, &color);

#warning Handle ROPs, etc. onto dst_buf here

		i++;
		src_idx += src_xinc;

		if ((i - ystart) == clipwidth) {
			int tmp;
			/* newline */

			/* do clipping here */
			i += (width - clipwidth);
			ystart = i;
			y++;

			LIB_ASSERT(((src_idx - src_ystart) >> SHIFT) >= src->w);

			tmp = src_y;
			src_y += src_yinc;

			if ((tmp >> SHIFT) == (src_y >> SHIFT)) {
				src_idx = src_ystart;
			} else {
				src_ystart = src_idx;
			}	/* if */
		}	/* if */

	}	/* for */

	return GALLOC_OK;
}	/* stretch_crossblit */


int stretch_multicrossblit(ggiBlt_source_list src, size_t nmemb,
			   ggiBlt_t blt, int width, int height)
{
	if (src == NULL) return GGI_EARGINVAL;
	if (nmemb == 0) return GGI_EARGINVAL;
	if (blt == NULL) return GGI_EARGINVAL;

	/* not implemented yet */
	return -1;
}	/* stretch_multicrossblit */


int stretch_polycrossblit(ggiBlt_source_list src, size_t nmemb,
			  ggiBlt_t blt, int width, int height)
{
	if (src == NULL) return GGI_EARGINVAL;
	if (nmemb == 0) return GGI_EARGINVAL;
	if (blt == NULL) return GGI_EARGINVAL;


	/* not implemented yet */
	return -1;
}	/* stretch_polycrossblit */







int stretch_crossvis2vis(ggi_visual_t src, int sx, int sy, int sw, int sh,
			 ggi_visual_t dst, int dx, int dy,
			 int width, int height, enum ggiGA_rop rop,
			 enum ggiGA_coordbase cm)
{
	int i;
	int y,ystart;

	fixed src_ystart;
	fixed src_idx, src_y;
	fixed src_xinc, src_yinc;

	ggi_color color;
	ggi_pixel *src_line;
	ggi_pixel *dst_line;


	if (src == NULL) return GGI_EARGINVAL;
	if (dst == NULL) return GGI_EARGINVAL;

#warning Convert dx,dy and width,height in pixels, if they are in dots


	src_line = malloc(width * height * sizeof(ggi_pixel));
	if (!src_line) return GGI_ENOMEM;
	dst_line = malloc(width * height * sizeof(ggi_pixel));
	if (!dst_line) {
		free(src_line);
		return GGI_ENOMEM;
	}	/* if */


	ystart = y = 0;


	src_ystart = src_y = 0;
	src_idx = 0;
	src_xinc = ((sw << SHIFT) / width) + 1;
	src_yinc = ((sh << SHIFT) / height) + 1;

	/* read first line */
	ggiGetHLine(src, sx, sy, sw, src_line);

	for (i = 0; i <= ((width - 1) * (height - 1)); ) {

		ggiUnmapPixel(src,
			src_line[((src_idx - src_ystart) >> SHIFT)],
			&color);
		dst_line[i - ystart] = ggiMapColor(dst, &color);

		i++;
		src_idx += src_xinc;

		if ((i - ystart) == width) {
			/* newline */

#warning Handle ROPs, clipping, etc. here

			ggiPutHLine(dst, dx, dy + y, width, dst_line);

			ystart = i;
			y++;

			LIB_ASSERT(((src_idx - src_ystart) >> SHIFT) >= sw);

			src_ystart = src_idx;
			src_y += src_yinc;

			/* read next line */
			ggiGetHLine(src, sx,sy + (src_y >> SHIFT), sw, src_line);
		}	/* if */
	}	/* for */

	free(src_line);
	free(dst_line);

	return GALLOC_OK;
}	/* stretch_crossvis2vis */


int stretch_crossblit2vis(ggiBlt_t blt, ggi_visual_t vis,
			  int dx, int dy, int width, int height,
			  enum ggiGA_rop rop, enum ggiGA_coordbase cm)
{
	if (blt == NULL) return GGI_EARGINVAL;
	if (vis == NULL) return GGI_EARGINVAL;

	if (vis == blt->vis) {
		if (cm != blt->cm) {

#warning Convert dx,dy, width and height into blt->cm
		}	/* if */

		return stretch_blit2vis(blt, dx,dy, width,height, rop);
	}	/* if */

	/* not finished yet */
	return -1;
}	/* stretch_crossblit2vis */


int stretch_blit2vis(ggiBlt_t blt, int dx, int dy, int width, int height,
		   enum ggiGA_rop rop)
{
	struct GA_x_bob *bob = NULL;
	enum ggiGA_coordbase cm;

	int i;
	int y,ystart;

	fixed bob_ystart;
	fixed bob_idx, bob_y;
	fixed bob_xinc, bob_yinc;

	ggi_color color;
	ggi_pixel *bob_buf;
	ggi_pixel *buf_line;


	if (blt == NULL) return GGI_EARGINVAL;

	cm = blt->cm;

#warning Convert dx,dy and width,height in pixels, if they are in dots


	if (width == 0 || height == 0) {
		/* nothing to do, because a zero-line is NOT visible */
		return GALLOC_OK;
	}	/* if */

	bob = GETBOB(blt);
	bob_buf = bob->bobimage;

	buf_line = malloc(width * height * sizeof(ggi_pixel));
	if (!buf_line) return GGI_ENOMEM;


	ystart = y = 0;

	bob_ystart = bob_y = 0;
	bob_idx = 0;
	bob_xinc = ((blt->w << SHIFT) / width) + 1;
	bob_yinc = ((blt->h << SHIFT) / height) + 1;

	for (i = 0; i <= ((width - 1) * (height - 1)); ) {
		ggi_pixel pix_tmp;
		int x;

		x = i - ystart;
		ggiUnmapPixel(blt->vis, bob_buf[(bob_idx >> SHIFT)], &color);
		pix_tmp = ggiMapColor(blt->vis, &color);

		if (blt->alpha != NULL) {
			unsigned char *alpha_ptr;
			unsigned char alpha_val;

			alpha_ptr = (unsigned char *)blt->alpha;
			alpha_val = alpha_ptr[(bob_idx >> SHIFT)];

			switch(alpha_val) {
			case 0:		/* Opaque */
				buf_line[x] = pix_tmp;
				break;
			case 0xff:	/* Transparent */
				ggiGetPixel(blt->vis, dx + x, dy + y,
					&buf_line[x]);
				break;
			default:	/* Partially Transparent*/
				{
				ggi_color sprcol,bgcol;
				ggi_pixel bgpix;

				ggiGetPixel(blt->vis, dx + x, dy + y,
					    &bgpix);
				ggiUnmapPixel(blt->vis, buf_line[x], &sprcol);
				ggiUnmapPixel(blt->vis, bgpix, &bgcol);

				sprcol.r = (sprcol.r * (255 - alpha_val)
					+ bgcol.r * alpha_val) / 255;
				sprcol.g = (sprcol.g * (255 - alpha_val)
					+ bgcol.g * alpha_val) / 255;
				sprcol.b = (sprcol.b * (255 - alpha_val)
					+ bgcol.b * alpha_val) / 255;

				buf_line[x] = ggiMapColor(blt->vis, &sprcol);
				}
                        }       /* switch */
		} else {
			buf_line[x] = ggiMapColor(blt->vis, &color);
		}	/* if */

		i++;
		x++;
		bob_idx += bob_xinc;


		LIB_ASSERT((i - ystart) <= width);

		if (x == width) {
			int tmp;
			/* newline */

#warning Handle ROPs, etc. here

			/* Draw the line.
			 * ggiPutHLine performs clipping, too.
			 */
			ggiPutHLine(blt->vis, dx, dy + y, width, buf_line);

			ystart = i;
			y++;

			LIB_ASSERT(((bob_idx - bob_ystart) >>SHIFT) >= blt->w);

			tmp = bob_y;
			bob_y += bob_yinc;

			if ((tmp >> SHIFT) == (bob_y >> SHIFT)) {
				bob_idx = bob_ystart;
			} else {
				bob_ystart = bob_idx;
			}	/* if */

		}	/* if */

	}	/* for */

	free(buf_line);

	return GALLOC_OK;
}	/* stretch_blit2vis */
